/* This file is NOT part of KDE ;)
   Copyright (C) 2006 Thomas Lbking <thomas.luebking@web.de>

   This application is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#ifdef QT4

// #include <QtGui>
// #include <QtOpenGL>
#include <QHBoxLayout>
#include <QFrame>
#include <QMenuBar>
#include <QMenu>
#include <QApplication>
#include <QInputEvent>
#include <QStringList>
#include <QTimer>
#include <QDateTime>
#include <QDebug>

#else
// #include <qpushbutton.h>
// #include <qslider.h>
#include <qlayout.h>
#include <qframe.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qapplication.h>
// #include <qkeycode.h>
#include <qstringlist.h>
#include <qtimer.h>
#include <qdatetime.h>

#endif

#include "kgliv.h"

#ifndef CLAMP
#define CLAMP(x,l,u) (x) < (l) ? (l) :\
(x) > (u) ? (u) :\
(x)
#endif

#ifndef QMAX
#define QMAX(x,y) (x) > (y) ? (x) : (y)
#endif

/**
Ok, hello and welcome to al little KGlImageView demo app.
This is not a maintaned app but only a short tutorial for using the KGLImageViewer.
Please don't mourn about functional bugs - THIS DEMO IS NOT MAINTAINED
*******************************************************/

static GLhandleARB color_vert, color_frag;

/** Just building the app framework - just some Qt code, nothing GL special... */
KGLIV::KGLIV( QWidget* parent, const char* name )
#ifdef QT4
: QWidget( parent )
#else
: QWidget( parent, name )
#endif
{
#ifdef QT4
   setAccessibleName( name );
#endif
   prev = current = next = 0L;
   
   _updated = true;
   _message = true;
   // Create nice frames to put around the OpenGL widgets
   QFrame* f1 = new QFrame( this );
   f1->setFrameStyle( QFrame::NoFrame );

   files = QStringList();
   counter = 0;
   for ( int i = 1; i < qApp->argc(); i++ )
      files += qApp->argv()[i];

   // Create an OpenGL widget
   view = new KGLImageView::KGLImageViewer( f1, "glbox1");
   if (view->providesShaders())
   {
      qWarning("trying to use shaders");
      char * cv = getenv("COLOR_VERTEX_SHADER");
      if (!qstrcmp( cv, "" ))
         cv = "color.vert";
      char * cp = getenv("COLOR_PIXEL_SHADER");
      if (!qstrcmp( cp, "" ))
         cp = "color.frag";
      color_vert = view->loadShader(cv, GL_VERTEX_SHADER_ARB);
      color_frag = view->loadShader(cp, GL_FRAGMENT_SHADER_ARB);
   }
//    view->setCanvas(Qt::green);
   QFont fnt = view->font();
   fnt.setPointSize(20);
   view->setFont(fnt);

#ifdef QT4
   view->setFocusPolicy( Qt::WheelFocus );
#else
   view->setFocusPolicy( QWidget::WheelFocus );
#endif
   view->installEventFilter(this);
//    view->setPaletteForegroundColor(Qt::black);
   load(true);

   current = &view->images().last();
//    current->setClipRect(200,300,400,600);
//    current->tint(QColor(255,0,0));
//    current->setBrightness(0, 5000);
//    QTimer::singleShot(500, this, SLOT(resetView()));
   if (files.count() > 1)
   {
      ++counter;
      load(false);
      next = &view->images().last();
      --counter;
   }
   if (_message)
      view->message(20,20,"\"" + files[counter] + "\"",5000);
   // Create a menu
#ifdef QT4
   QMenu *file = new QMenu( "&File", this );
   file->addAction( "Exit",  qApp, SLOT(quit()), Qt::CTRL+Qt::Key_Q );
   
   QMenu *edit = new QMenu( "&View", this );
   edit->addAction( "Toggle Fullscreen mode", this, SLOT(toggleFullscreen()), Qt::ALT+Qt::Key_Return );
   edit->addAction( "Toggle Filename Display", this, SLOT(toggleMessage()), Qt::Key_M );
//    edit->addAction( "Toggle Freeze", this, SLOT(toggleFreeze()), Qt::Key_F );
   edit->addAction( "Reset View", this, SLOT(resetView()), Qt::Key_Asterisk );
   edit->addAction( "Maximize", this, SLOT(maxW()), Qt::Key_Slash );
   edit->addAction( "zOOm IN", this, SLOT(zoomIn()), Qt::Key_Plus );
   edit->addAction( "zOOm out", this, SLOT(zoomOut()), Qt::Key_Minus );
   edit->addAction( "Blur more", this, SLOT(blurMore()), Qt::SHIFT+Qt::Key_S );
   edit->addAction( "Blur less", this, SLOT(blurLess()), Qt::Key_S );
   edit->addAction( "Tunnel more", this, SLOT(tunnelMore()), Qt::SHIFT+Qt::Key_T );
   edit->addAction( "Tunnel less", this, SLOT(tunnelLess()), Qt::Key_T );
   
   QMenu *rotation = new QMenu( "&Rotation", this );
   rotation->addAction( "90 CW", this, SLOT(r90cw()), Qt::CTRL+Qt::Key_Right );
   rotation->addAction( "90 CCW", this, SLOT(r90ccw()), Qt::CTRL+Qt::Key_Left );
   rotation->addAction( "Flip Horizontal 'CW'", this, SLOT(fhcw()), Qt::CTRL+Qt::Key_Down );
   rotation->addAction( "Flip Horizontal 'CCW'", this, SLOT(fhccw()), Qt::CTRL+Qt::Key_Up );
   rotation->addAction( "Flip Vertical 'CW'", this, SLOT(fvcw()), Qt::CTRL+Qt::SHIFT+Qt::Key_Right );
   rotation->addAction( "Flip Vertical 'CCW'", this, SLOT(fvccw()), Qt::CTRL+Qt::SHIFT+Qt::Key_Left );
   
   QMenu *colors = new QMenu( "&Colors", this );
   colors->addAction( "Increase Brightness", this, SLOT(brighter()), Qt::SHIFT+Qt::Key_L );
   colors->addAction( "Decrease Brightness", this, SLOT(darker()), Qt::Key_L );
   colors->addAction( "Increase Red", this, SLOT(redUp()), Qt::SHIFT+Qt::Key_R );
   colors->addAction( "Decrease Red", this, SLOT(redDown()), Qt::Key_R );
   colors->addAction( "Increase Green", this, SLOT(greenUp()), Qt::SHIFT+Qt::Key_G );
   colors->addAction( "Decrease Green", this, SLOT(greenDown()), Qt::Key_G );
   colors->addAction( "Increase Blue", this, SLOT(blueUp()), Qt::SHIFT+Qt::Key_B );
   colors->addAction( "Decrease Blue", this, SLOT(blueDown()), Qt::Key_B );
   colors->addAction( "Increase Alpha", this, SLOT(alphaUp()), Qt::SHIFT+Qt::Key_A );
   colors->addAction( "Decrease Alpha", this, SLOT(alphaDown()), Qt::Key_A );
   colors->addAction( "Decrease Alpha", this, SLOT(invert()), Qt::Key_I );
   colors->addAction( "Reset Colors", this, SLOT(resetColors()), Qt::Key_C );
   
   // Create a menu bar
   m = new QMenuBar( this );
   m->addMenu( file );
   m->addMenu( edit );
   m->addMenu( rotation );
   m->addMenu( colors );
   
   // Put the GL widget inside the frame
   QHBoxLayout* flayout1 = new QHBoxLayout( f1 );
   flayout1->setMargin(0);
   flayout1->setSpacing(0);
   flayout1->addWidget( view, 1 );
   
   // Top level layout, puts the sliders to the left of the frame/GL widget
   QHBoxLayout* hlayout = new QHBoxLayout( this );
   hlayout->setMargin(0);
   hlayout->setSpacing(0);
   hlayout->setMenuBar( m );
   hlayout->addWidget( f1, 1 );
#else
   QPopupMenu *file = new QPopupMenu( this );
   file->insertItem( "Exit",  qApp, SLOT(quit()), CTRL+Key_Q );
   
   QPopupMenu *edit = new QPopupMenu( this );
   edit->insertItem( "Toggle Fullscreen mode", this, SLOT(toggleFullscreen()), ALT+Key_Return );
   edit->insertItem( "Toggle Filename Display", this, SLOT(toggleMessage()), Key_M );
//    edit->insertItem( "Toggle Freeze", this, SLOT(toggleFreeze()), Key_F );
   edit->insertItem( "Reset View", this, SLOT(resetView()), Key_Asterisk );
   edit->insertItem( "Maximize", this, SLOT(maxW()), Key_Slash );
   edit->insertItem( "zOOm IN", this, SLOT(zoomIn()), Key_Plus );
   edit->insertItem( "zOOm out", this, SLOT(zoomOut()), Key_Minus );
   
   QPopupMenu *rotation = new QPopupMenu( this );
   rotation->insertItem( "90 CW", this, SLOT(r90cw()), CTRL+Key_Right );
   rotation->insertItem( "90 CCW", this, SLOT(r90ccw()), CTRL+Key_Left );
   rotation->insertItem( "Flip Horizontal 'CW'", this, SLOT(fhcw()), CTRL+Key_Down );
   rotation->insertItem( "Flip Horizontal 'CCW'", this, SLOT(fhccw()), CTRL+Key_Up );
   rotation->insertItem( "Flip Vertical 'CW'", this, SLOT(fvcw()), CTRL+SHIFT+Key_Right );
   rotation->insertItem( "Flip Vertical 'CCW'", this, SLOT(fvccw()), CTRL+SHIFT+Key_Left );
   
   QPopupMenu *colors = new QPopupMenu( this );
   colors->insertItem( "Increase Brightness", this, SLOT(brighter()), SHIFT+Key_L );
   colors->insertItem( "Decrease Brightness", this, SLOT(darker()), Key_L );
   colors->insertItem( "Increase Red", this, SLOT(redUp()), SHIFT+Key_R );
   colors->insertItem( "Decrease Red", this, SLOT(redDown()), Key_R );
   colors->insertItem( "Increase Green", this, SLOT(greenUp()), SHIFT+Key_G );
   colors->insertItem( "Decrease Green", this, SLOT(greenDown()), Key_G );
   colors->insertItem( "Increase Blue", this, SLOT(blueUp()), SHIFT+Key_B );
   colors->insertItem( "Decrease Blue", this, SLOT(blueDown()), Key_B );
   colors->insertItem( "Increase Alpha", this, SLOT(alphaUp()), SHIFT+Key_A );
   colors->insertItem( "Decrease Alpha", this, SLOT(alphaDown()), Key_A );
   colors->insertItem( "Reset Colors", this, SLOT(resetColors()), Key_C );
    
   // Create a menu bar
   m = new QMenuBar( this );
   m->insertItem("&File", file );
   m->insertItem("&View", edit );
   m->insertItem("&Rotation", rotation );
   m->insertItem("&Colors", colors );

   // Put the GL widget inside the frame
   QHBoxLayout* flayout1 = new QHBoxLayout( f1, 0, 0, "flayout1");
   flayout1->addWidget( view, 1 );

   // Top level layout, puts the sliders to the left of the frame/GL widget
   QHBoxLayout* hlayout = new QHBoxLayout( this, 0, 0, "hlayout");
   hlayout->setMenuBar( m );
   hlayout->addWidget( f1, 1 );
#endif

}

/** Resets the view */
void KGLIV::resetView()
{
   /// this sets the rotation of all 3 axes to zero. "500" is the amount of ms this action should last (1/2 sec)
   view->rotateTo(KGLImageView::X,0.0,500);
   view->rotateTo(KGLImageView::Y,0.0,500);
   view->rotateTo(KGLImageView::Z,0.0,500);
   
   view->scaleTo(KGLImageView::X, 100.0, 500);
   view->scaleTo(KGLImageView::Y, 100.0, 500);
   
   current->rotateTo(KGLImageView::X,0.0,500);
   current->rotateTo(KGLImageView::Y,0.0,500);
   current->rotateTo(KGLImageView::Z,0.0,500);
   current->resize(current->basicWidth(), current->basicHeight(), 500, 100.0, 100.0);
#if 0
   if (view->images().last().basicHeight() > view->images().last().basicWidth())
   {
      /** scaling the view to 100% on either axis will keep the aspec ratio, but try to fill the whole width
      as we have a free aspect ratio and want the whole (last) image to be visable, we must reduce the value by the image ratio in case it's higher than wide
      */
      float p = 100.0*view->images().last().basicWidth()/view->images().last().basicHeight();
      view->scaleTo(KGLImageView::X,p,500);
      view->scaleTo(KGLImageView::Y,p,500);
   }
   else
   {
      view->scaleTo(KGLImageView::X,100.0,500);
      view->scaleTo(KGLImageView::Y,100.0,500);
   }
#endif
   
   /// set the position on axes X and Y to zero (that is: center the view)
   view->moveTo(KGLImageView::X,50.0,500);
   view->moveTo(KGLImageView::Y,50.0,500);
   /** NOTICE the Z axis is set to -5.0. what can be interpreted as moving 5 steps backward.
   this value is arbitrary.
   the value should be smaller than -3.0, otherwise nothing placed at 0.0 (the images) will be visable
   (put a peace of paper right in front of your eyes and try to read sth. to understand this)
   the additional 2 steps are made so that the view doesn't get clipped as soon as it's rotated slightly by the X or Y axis ("into" the screen)
   */
   view->moveTo(KGLImageView::Z, 5.0,500);
}

/** rotate by 90 clockwise*/
void KGLIV::r90cw()
{
   view->rotate(KGLImageView::Z,90.0,500);
}

/** rotate by 90 counterclockwise*/
void KGLIV::r90ccw()
{
   view->rotate(Z,-90.0,500);
}

/** rotate by 180 clockwise around the X axis*/
void KGLIV::fhcw()
{
   view->rotate(KGLImageView::X,180.0,500);
}

/** rotate by 180 counter clockwise around the X axis*/
void KGLIV::fhccw()
{
   view->rotate(KGLImageView::X,-180.0,500);
}

/** rotate by 180 clockwise around the Y axis*/
void KGLIV::fvcw()
{
   view->rotate(KGLImageView::Y,180.0,500);
}

/** rotate by 180 counter clockwise around the Y axis*/
void KGLIV::fvccw()
{
   view->rotate(KGLImageView::Y,-180.0,500);
}

/** zoom (in) in by 125%*/
void KGLIV::zoomIn()
{
   view->scale(KGLImageView::X, 125.0,500);
   view->scale(KGLImageView::Y, 125.0,500);
}

/** zoom (out) in by 80% == 1/125%*/
void KGLIV::zoomOut()
{
   view->scale(KGLImageView::X, 80.0, 500);
   view->scale(KGLImageView::Y, 80.0, 500);
}

void KGLIV::brighter()
{
   float val = current->brightness()+5.0;
   current->setBrightness(val);
   view->message(20,20, QString("\"" + files[counter] + "\"\nBrightness: %1\%").arg(val), 1000);
}

void KGLIV::darker()
{
   float val = current->brightness()-5.0;
   current->setBrightness(val);
   view->message(20,20, QString("\"" + files[counter] + "\"\nBrightness: %1\%").arg(val), 1000);
}

#ifdef QT4
const char* blurstring = "\"\nBlurrage: %1x%1 kernel";
const char* tunnelstring = "\"\nTunnel: %1x%1 kernel";
#else
const char* blurstring = "\"\nBlurrage: %1%";
const char* tunnelstring = "\"\nTunnel: %1%";
#endif

void KGLIV::blurMore()
{
   float fact = current->blurrage()+0.5;
   current->boxBlur(fact);
   view->message(20,20, QString("\"" + files[counter] + blurstring).arg(2*fact+1), 1000);
}

void KGLIV::blurLess()
{
   float fact = QMAX(0,current->blurrage()-0.5);
   current->boxBlur(fact);
   if (fact)
      view->message(20,20, QString("\"" + files[counter] + blurstring).arg(2*fact+1), 1000);
   else
      view->message(20,20, QString("\"" + files[counter] + "\"\nBlurrage: off"), 1000);
}

void KGLIV::tunnelMore()
{
   float fact = current->blurrage()+0.5;
   current->tunnel(fact);
   view->message(20,20, QString("\"" + files[counter] + tunnelstring).arg(2*fact+1), 1000);
}

void KGLIV::tunnelLess()
{
   float fact = QMAX(0,current->blurrage()-0.5);
   current->tunnel(fact);
   if (fact)
      view->message(20,20, QString("\"" + files[counter] + tunnelstring).arg(2*fact+1), 1000);
   else
      view->message(20,20, QString("\"" + files[counter] + "\"\nTunnel: off"), 1000);
}

void KGLIV::changeColor(int key, int v)
{
   QColor c = current->color();
   float a = current->alpha();
   if (key == Qt::Key_A) // alpha
   {
      a = CLAMP(a+v,0.0,100.0);
      current->setAlpha(a,200);
   }
   else
   {
      if (key == Qt::Key_R)
         c = QColor(CLAMP(c.red()+v,0,255),c.green(),c.blue());
      else if (key == Qt::Key_G)
         c = QColor(c.red(),CLAMP(c.green()+v,0,255),c.blue());
      else if (key == Qt::Key_B)
         c = QColor(c.red(),c.green(),CLAMP(c.blue()+v,0,255));
      current->tint(c,200);
   }
   view->message(20,20, QString("R: %1, G: %2, B: %3, A: %4\%").arg(c.red()).arg(c.green()).arg(c.blue()).arg(a), 1000);
}

void KGLIV::resetColors()
{
   current->setAlpha(100.0, 500);
   current->tint(Qt::white, 500);
}

void KGLIV::invert()
{
   current->invert(!current->isInverted());
}

void KGLIV::redUp(){ changeColor(Qt::Key_R, 10); }
void KGLIV::redDown(){ changeColor(Qt::Key_R, -10); }
void KGLIV::greenUp(){ changeColor(Qt::Key_G, 10); }
void KGLIV::greenDown(){ changeColor(Qt::Key_G, -10); }
void KGLIV::blueUp(){ changeColor(Qt::Key_B, 10); }
void KGLIV::blueDown(){ changeColor(Qt::Key_B, -10); }
void KGLIV::alphaUp(){ changeColor(Qt::Key_A, 5); }
void KGLIV::alphaDown(){ changeColor(Qt::Key_A, -5); }


void KGLIV::maxW()
{
   view->scaleTo(KGLImageView::X, 100.0, 500);
   view->scaleTo(KGLImageView::Y, 100.0, 500);
   view->rotateTo(KGLImageView::X,0.0,500);
   view->rotateTo(KGLImageView::Y,0.0,500);
   view->rotateTo(KGLImageView::Z,0.0,500);
   if ((float)view->height()/current->basicHeight() > (float)view->width()/current->basicWidth())
      current->scaleTo(KGLImageView::Y, current->scaleTo(KGLImageView::X, 100.0, 500, true, 100.0), 500);
   else
      current->scaleTo(KGLImageView::X, current->scaleTo(KGLImageView::Y, 100.0, 500, true, 100.0), 500);
   view->moveTo(KGLImageView::X,50.0,500);
   view->moveTo(KGLImageView::Y,50.0,500);
}

void KGLIV::updateCache()
{
   if (next == 0 && counter < files.count()-1)
   {
//       prev->hide();
      ++counter;
      load(false);
      --counter;
      next = &view->images().last();
   }
   else if (prev == 0 && counter > 0)
   {
//       next->hide();
      --counter;
      load(false);
      ++counter;
      prev = &view->images().last();
   }
   _updated = true;
}

void KGLIV::nextImage()
{
   KGLImage *oldImage = current;
   current->hide();
   if (prev == 0)
   {
      prev = current;
      current = next;
      next = 0L;
   }
   else if (next == 0)
   {
      next = current;
      current = prev;
      prev = 0L;
   }
   if (current->hasAlpha())
      view->setCanvas(Qt::gray);
   else
      view->setCanvas(Qt::black);
   switch (effect)
   {
   case 0:
      oldImage->setAlpha(100.0,false);
      current->move(KGLImageView::Z, 0.001);
      break;
   case 1:
      oldImage->setAlpha(100.0,false);
      current->show(false);
      current->setAlpha(100.0,300);
      break;
   case 2:
      oldImage->rotateTo(0,0,0,false);
      current->show(false);
      current->rotate(KGLImageView::Y, sign*90.0,300);
      break;
   case 3:
   default:
      oldImage->rotateTo(0,0,0,false);
      current->show(false);
      current->rotate(KGLImageView::X, sign*90.0,300);
   }
   resetView();
   QTimer::singleShot(600, this, SLOT(updateCache()));
//    view->images().last().tint(Qt::green,3000);
   if (_message)
      view->message(20,20,"\"" + files[counter] + "\"",5000);
#ifdef QT4
   setWindowTitle ( files[counter] );
#else
   setCaption(files[counter]);
#endif
}

void KGLIV::load(bool show)
{
   while (view->load(files[counter], show) < 0)
   {
#ifdef QT4
      files.removeAt(counter);
#else
      files.remove(files.at(counter));
#endif
      if (files.count() < counter+1)
         counter = files.count() - 1;
      if (files.count() <= 0)
         qApp->quit();
   }
   view->images().last().addShader(color_vert, false);
   view->images().last().addShader(color_frag);
}

void KGLIV::toggleFullscreen()
{
   if (isFullScreen())
   {
      showNormal();
      m->show();
   }
   else
   {
      m->hide();
      showFullScreen();
   }
}

void KGLIV::toggleMessage()
{
   if (_message)
      view->hideMessage();
   _message = !_message;
}


/** handle the user input
* this function is a horrible mess - don't mourn.
* the interesting part is documented slight below - the rest is just key/mousebutton mapping, nothing special
*/

#ifdef QT4
#define _ALT_ Qt::AltModifier
#define _SHIFT_ Qt::ShiftModifier
#define _CTRL_ Qt::ControlModifier
#else
#define _ALT_ Qt::AltButton
#define _SHIFT_ Qt::ShiftButton
#define _CTRL_ Qt::ControlButton
#endif

bool KGLIV::eventFilter ( QObject * o, QEvent * e )
{
//    if (e->type() != QEvent::Resize)
//       return false;
//    ((KGLImageViewer*)o)->images().last().scaleTo(100.0, 100.0, true, true);
   int direction = 0;
#ifdef QT4
   Qt::KeyboardModifiers state = ((QInputEvent*)e)->modifiers();
#endif
   if (e->type() == QEvent::KeyPress)
   {
#ifndef QT4
      Qt::ButtonState state = ((QKeyEvent*)e)->state();
#endif
      switch (((QKeyEvent*)e)->key())
      {
      case Qt::Key_Enter:
      case Qt::Key_Return:
         if ( state & _ALT_)
            toggleFullscreen(); return false;
      case Qt::Key_PageUp:
         direction = -1; break;
      case Qt::Key_PageDown:
         direction = 1; break;
      case Qt::Key_Q:
         if (state & _CTRL_)
            qApp->quit(); return false;
      case Qt::Key_Asterisk:
         resetView(); return false;
      case Qt::Key_Slash:
         maxW(); return false;
      case Qt::Key_Plus:
         zoomIn(); return false;
      case Qt::Key_Minus:
         zoomOut(); return false;
      case Qt::Key_Right:
         if (state & _SHIFT_ && state & _CTRL_)
            fvcw();
         else if (state & _CTRL_)
            r90cw();
         return false;
      case Qt::Key_Left:
         if (state & _SHIFT_ && state & _CTRL_)
            fvccw();
         else if (state & _CTRL_)
            r90ccw();
         return false;
      case Qt::Key_Down:
         if (state & _CTRL_)
            fhcw(); return false;
      case Qt::Key_Up:
         if (state & _CTRL_)
            fhccw(); return false;
      case Qt::Key_M:
         toggleMessage(); return false;
      case Qt::Key_L:
         if (state & _SHIFT_)
            brighter();
         else
            darker();
         return false;
      case Qt::Key_A:
      case Qt::Key_R:
      case Qt::Key_G:
      case Qt::Key_B:
      {
         int key = ((QKeyEvent*)e)->key();
         if (state & _SHIFT_)
            changeColor(key, key == Qt::Key_A ? 5 : 10);
         else
            changeColor(key, key == Qt::Key_A ? -5 : -10);
         return false;
      }
      case Qt::Key_S:
      {
         if (state & _SHIFT_)
            blurMore();
         else
            blurLess();
         return false;
      }
      case Qt::Key_T:
      {
         if (state & _SHIFT_)
            tunnelMore();
         else
            tunnelLess();
         return false;
      }
      case Qt::Key_C:
         resetColors(); return false;
      case Qt::Key_I:
         invert(); return false;
      default:
         return false;
      }
   }
   else if (e->type() == QEvent::Wheel)
   {
      if (((QWheelEvent*)e)->delta() < 0 )
         direction = 1;
      else
         direction = -1;
   }
   else
      return false;
   if (!_updated)
      return false;
   KGLImage *newImg;
   if ( direction == 1 )
   {
      if (counter == files.count()-1)
         return false;
      if (prev)
      {
         view->remove(prev->id());
         prev = 0L;
      }
      newImg = next;
      ++counter;
      sign = 1;
   }
   else if (direction == -1 && counter > 0)
   {
      if (next)
      {
         view->remove(next->id());
         next = 0L;
      }
      newImg = prev;
      --counter;
      sign = -1;
   }
   else
      return false;
   _updated = false;
   effect = QDateTime::currentDateTime().toTime_t() % 4;
   switch (effect)
   {
   case 0: // crossfade
      newImg->setAlpha(0.0);
      newImg->move(KGLImageView::Z, -0.001);
      newImg->show(false);
      current->setAlpha(0.0,500);
      newImg->setAlpha(100.0,500);
      QTimer::singleShot(501, this, SLOT(nextImage()));
      break;
   case 1: // fade out/in
      newImg->setAlpha(0.0);
      current->setAlpha(0.0,300);
      QTimer::singleShot(301, this, SLOT(nextImage()));
      break;
   case 2: // rotate horizontally
      current->rotate(KGLImageView::Y, sign*90.0,300);
      newImg->rotate(KGLImageView::Y, sign*-90.0);
      QTimer::singleShot(301, this, SLOT(nextImage()));
      break;
   case 3: // rotate vertically
   default:
      current->rotate(KGLImageView::X, sign*90.0,300);
      newImg->rotate(KGLImageView::X, sign*-90.0);
      QTimer::singleShot(301, this, SLOT(nextImage()));
   }
   return false;
}
